/* socketpair.h
Copyright 2007, 2010 by Nathan C. Myers <ncm@cantrip.org>
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    The name of the author must not be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SOCKETPAIR_H
#define SOCKETPAIR_H

#include "base/visibility.hpp"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifdef I2_SOCKETPAIR_BUILD
#       define I2_SOCKETPAIR_API I2_EXPORT
#else
#       define I2_SOCKETPAIR_API I2_IMPORT
#endif /* I2_SOCKETPAIR_BUILD */

#ifdef _WIN32
I2_SOCKETPAIR_API int dumb_socketpair(SOCKET socks[2], int make_overlapped);
#else /* _WIN32 */
I2_SOCKETPAIR_API int dumb_socketpair(int socks[2], int dummy);
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SOCKETPAIR_H */

